import random

import torch
import numpy as np
from torch import nn
from visdom import Visdom

class AverageMeter(object):
    """Keeps track of most recent, average, sum, and count of a metric.

    Example
    -------
    losses = AverageMeter()
    losses.update(1, 5)
    print(losses.avg)
    """

    def __init__(self):
        self.reset()

    def reset(self):
        # Reset all value to 0.
        self.value = 0
        self.avg   = 0
        self.sum   = 0
        self.count = 0

    def update(self, value, n=1):
        """Update value, average, sum, and count.

        Parameters
        ----------
        n : int, optional (default = 5)
        value : double

        """
        self.value = value
        self.sum   += value * n
        self.count += n
        self.avg   = self.sum / self.count

class VisdomLinePlotter(object):
    """Plots lines in Visdom.

    Parameter
    ----------
    env_name: str, optional (default = 'main')
        Environment name of Visdom, you should not change it if don't know what's going on.

    Example
    -------
    import time

    plotter = VisdomLinePlotter()
    for x, y in zip(range(10), range(10)):
        plotter.plot("var_name", "split_name", "title_name", x, y)
        time.sleep(2)
    """
    def __init__(self, env_name='main'):
        self.viz = Visdom()
        self.env = env_name
        self.plots = {}

    def plot(self, var_name, split_name, title_name, x, y):
        if var_name not in self.plots:
            self.plots[var_name] = self.viz.line(X   = np.array([x,x]),
                                                 Y   = np.array([y,y]),
                                                 env = self.env,
                                                 opts= dict(legend = [split_name],
                                                           title   = title_name,
                                                           xlabel  = 'Epochs',
                                                           ylabel  = var_name))

        else:
            self.viz.line(X      = np.array([x]),
                          Y      = np.array([y]),
                          env    = self.env,
                          win    = self.plots[var_name],
                          name   = split_name,
                          update = 'append')

def pytorch_reproducer(device="cpu", seed=2019):
    """Reproducer for pytorch experiment.

    Parameters
    ----------
    seed: int, optional (default = 2019)
    	Radnom seed.
    device: str, optinal (default = "cpu")
    	Device type.

    Example
    -------
    pytorch_reproducer(seed=2019, device=DEVICE).
    """
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    if device == 'cuda':
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True

def get_device():
    """Return device type.

    Return
    ------
    DEVICE: torch.device

    Example
    -------
    DEVICE = get_device()
    """
    DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    return DEVICE

def clip_gradient(optimizer, grad_clip):
    """Clip gradients computed during backpropagation to prevent gradient explosion.

     Parameters
     ----------
     optimizer: pytorch optimizer
     	Optimized with the gradients to be clipped.
     grad_clip: double
     	Gradient clip value.

     Example
     -------
     from torch.optim import Adam
     from torchvision import models

     model = models.AlexNet()
     optimizer = Adam(model.parameters())
     clip_gradient(optimizer, 5)
     """
    for group in optimizer.param_groups:
        for param in group['params']:
            if param.grad is not None:
                param.grad.data.clamp_(-grad_clip, grad_clip)

def adjust_learning_rate(optimizer, scale_factor):
    """Shrinks learning rate by a specified factor.

    Parameters
    ----------
    optimizer: pytorch optimizer
    scale_factor: factor to scale by
    """

    print("\nDECAYING learning rate.")
    for param_group in optimizer.param_groups:
        param_group['lr'] = param_group['lr'] * scale_factor
    print("The new learning rate is %f\n" % (optimizer.param_groups[0]['lr'],))

def get_learning_rate(optimizer):
    """Get learning rate.

    Parameters
    ----------
    optimizer: pytorch optimizer
    """
    lr=[]
    for param_group in optimizer.param_groups:
       lr +=[ param_group['lr'] ]

    assert(len(lr)==1) #we support only one param_group
    lr = lr[0]

    return lr

def init_conv2d(m):
    """Init parameters of convolution layer(初始化卷积层参数)

       Parameters
       ----------
       m: pytorch model

       Example
       -------
       class model(nn.Module):

           def __init__(self):
               super().__init__()

               # 初始化卷积层权重
               init_conv2d(self)
    """
    # 遍历网络子节点
    for c in m.modules():
        # 初始化卷积层
        if isinstance(c, nn.Conv2d):
            nn.init.xavier_uniform_(c.weight)
            nn.init.kaiming_normal_(c.weight, mode='fan_out', nonlinearity='relu')
            if c.bias is not None:
                nn.init.constant_(c.bias, 0.)
        # 初始BatchNorm层
        elif isinstance(c, nn.BatchNorm2d):
            nn.init.constant_(c.weight, 1.)
            nn.init.constant_(c.bias, 0.)
        # 初始线性层
        elif isinstance(c, nn.Linear):
            nn.init.normal_(c.weight, 0., 0.01)
            nn.init.constant_(c.bias, 0.)

def decimate(tensor, dims):
    """将tensor的维度变为dims
       Parameters
       ----------
       tensor: pytorch tensor
       dims: list

       Example
       -------
       x = torch.rand(4096, 512, 7, 7)
       decimate(x, [1024, None, 3, 3])
    """
    assert tensor.dim() == len(dims)

    for i in range(len(dims)):
        if dims[i] is not None:
            tensor = tensor.index_select(dim=i, index=torch.linspace(0, tensor.size()[i]-1, dims[i]).long())

    return tensor



# from ..utils.KA_utils import tick_tock
# import numpy as np
# import pandas as pd
#
#
# def ka_remove_duplicate_cols(df, **kwargs):
#     '''Remove duplicate columns
#
#        Parameters
#        ----------
#        df: pandas dataframe
#           Features matrix
#
#        **kwargs: all parameters in drop_duplicates function
#            subset : column label or sequence of labels, optional
#                 Only consider certain columns for identifying duplicates, by
#                 default use all of the columns
#            keep : {'first', 'last', False}, default 'first'
#                - ``first`` : Drop duplicates except for the first occurrence.
#                - ``last`` : Drop duplicates except for the last occurrence.
#                - False : Drop all duplicates.
#            take_last : deprecated
#            inplace : boolean, default False
#                 Whether to drop duplicates in place or to return a copy
#        Return
#        ------
#        new pandas dataframe with "unique columns" and "removed column names"
#
#        Example
#        -------
#        data_1_unique, removed_cols = ka_remove_duplicate_cols(data_1[numeric_cols])
#     '''
#     df_unique_columns = df.T.drop_duplicates(**kwargs).T
#     return df_unique_columns, set(df.columns.tolist()) - set(df_unique_columns.columns.tolist())



#
# from ..utils.k_others import tick_tock, callbacks_keras, LabelEncoder
#
# import lightgbm
# from lightgbm import callback
#
# import numpy as np
# import pandas as pd
# from tqdm import tqdm, tqdm_notebook
# from sklearn.svm import SVC, SVR
# from sklearn.utils import shuffle
# from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
# from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor
# from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
# from sklearn.linear_model import LogisticRegression, ElasticNet, Lasso, Ridge
# from sklearn.ensemble import GradientBoostingRegressor, GradientBoostingClassifier
# from sklearn.ensemble import AdaBoostClassifier, AdaBoostRegressor, ExtraTreesClassifier, ExtraTreesRegressor
#
#
#
# class ka_stacking_generalization(object):
#     def __init__(self, X_train, X_test, y_train, kf_n, verbose=1):
#         '''Stacking for models except "neuron network" and "xgboost"
#
#            Parameters
#            ----------
#            X_train: numpy array
#                 training data
#            X_test: numpy array
#                 testing data
#            y_train: numpy array
#                 training target
#            kf_n: KFold object from model_selection log
#
#            Example
#            -------
#            kf_5 = KFold(n_splits=5, random_state=413, shuffle=True)
#            stack_generater = ka_stacking_generalization(X.values, X_test.values, y.values, kf_5)
#         '''
#         self.X_train = X_train
#         self.X_test = X_test
#         self.y_train = y_train
#         self.kf_n = kf_n
#         self.verbose = verbose
#         self.cv_info = {'lgbm_info':{'cv_scores':[], 'cv_rounds':[], 'cv_losses':[]}}
#
#     def run_lgbm_stacker(self
#                          , lgbm_params
#                          , num_boost_round
#                          , early_stopping_rounds
#                          , lightgbm_verbose_eval
#                          , score_metric):
#         '''Stacking for xgboost
#
#            Parameters
#            ----------
#            lgbm_params: python dictionary
#                 lightgbm parameters
#            num_boost_round: int
#                 number of boosting rounds
#                 use as large as possible to overfitting, then decrese to see let CV and LB consistent.
#                 check difference number like 50 100 to see if they have big difference in CV, if have, one
#                 should notive underfitting
#            early_stopping_rounds: int
#                 number of early stopping round, if we do not want to use earlying stopping,
#                 set ---> early_stopping_rounds > num_boost_round
#            lightgbm_verbose_eval: 0 1 2
#                 verbose parameter in lightgbm
#            Return
#            ------
#            S_train: numpy array
#                             stacked training data
#            S_test: numpy array
#                             stacked testing data
#            Example
#            -------
#            params = {'application': 'binary'
#                      ,'learning_rate': 0.05
#                      ,'metric': 'auc'
#                      ,'feature_fraction': 0.8
#                      ,'bagging_fraction': 0.8
#                      ,'bagging_freq': 2
#                      ,'num_leaves': 80
#                      ,'max_depth': -1
#                      ,'lambda_l2':5
#                      ,'min_data_in_leaf': 10}
#
#            S_train, S_test, cv_scores = stack_generater.run_lgbm_stacker(params, 2000, 30, 0, roc_auc_score)
#         '''
#         with tick_tock("stacking", self.verbose):
#             n_folds = self.kf_n.n_splits
#             self.cv_info['lgbm_info'] = {'cv_scores':[], 'cv_rounds':[], 'cv_losses':[]}
#             S_train = np.zeros_like(self.y_train).astype(np.float32)
#             S_test_i = np.zeros((self.X_test.shape[0], n_folds)).astype(np.float32)
#
#             for i, (train_index, val_index) in enumerate(self.kf_n.split(self.X_train, self.y_train)):
#                 losses = dict()
#                 X_train_cv, X_valid_cv = self.X_train[train_index], self.X_train[val_index]
#                 y_train_cv, y_valid_cv = self.y_train[train_index], self.y_train[val_index]
#
#                 d_train_fold = lightgbm.Dataset(X_train_cv, y_train_cv)
#                 d_val_fold = lightgbm.Dataset(X_valid_cv, y_valid_cv)
#
#                 model_fold = lightgbm.train(lgbm_params
#                                             , train_set=d_train_fold
#                                             , valid_sets=d_val_fold
#                                             , early_stopping_rounds=early_stopping_rounds
#                                             , verbose_eval=lightgbm_verbose_eval
#                                             , num_boost_round=num_boost_round
#                                             , callbacks=[lightgbm.callback.record_evaluation(losses)])
#                 pred_valid = model_fold.predict(X_valid_cv)
#
#                 S_train[val_index] = pred_valid
#                 S_test_i[:, i] = model_fold.predict(self.X_test)
#
#                 score_tmp = score_metric(y_valid_cv, pred_valid)
#                 self.cv_info['lgbm_info']['cv_scores'].append(score_tmp)
#                 self.cv_info['lgbm_info']['cv_rounds'].append(model_fold.best_iteration)
#                 self.cv_info['lgbm_info']['cv_losses'].append(losses[list(losses.keys())[0]][list(losses[list(losses.keys())[0]].keys())[0]])
#                 print("Fold:{} --> score:{}.".format(i, score_tmp))
#
#             S_test = S_test_i.sum(axis=1) / n_folds
#             print("Mean:{}, Std:{}".format(np.mean(self.cv_info['lgbm_info']['cv_scores'])
#                                            , np.std(self.cv_info['lgbm_info']['cv_scores'])))
#
#             return S_train, S_test
#
#     def plot_loss_curve(self, ax, info_type):
#         '''
#         _, ax = plt.subplots(nrows=3, ncols=2, figsize=[15,10])
#         info_type: 'lgbm_info'
#         '''
#         pic_ith = 0
#         for i in range(np.int(self.kf_n.n_splits / 2) + 1):
#             for j in range(2):
#                 if((i+1)*(j+1) > self.kf_n.n_splits):
#                     continue
#                 ax[i,j].set_title('round: ' + str(self.cv_info['lgbm_info']['cv_rounds'][pic_ith]) + \
#                           '  --  loss: ' + str(self.cv_info['lgbm_info']['cv_losses'][pic_ith][-1]))
#                 ax[i,j].plot(self.cv_info[info_type]['cv_losses'][pic_ith])
#                 pic_ith += 1
#
#     def run_xgboost_stacker(self
#                             , xgb_params
#                             , num_boost_round
#                             , early_stopping_rounds
#                             , xgboost_verbose_eval
#                             , score_metric
#                             , verbose=1):
#         with tick_tock("stacking", self.verbose):
#             d_test = xgboost.DMatrix(self.X_test)
#             n_folds = self.kf_n.n_splits
#             S_train = np.zeros_like(self.y_train).astype(np.float32)
#             S_test_i = np.zeros((self.X_test.shape[0], n_folds)).astype(np.float32)
#             cv_scores = []
#
#             for i, (train_index, val_index) in enumerate(self.kf_n.split(self.X_train, self.y_train)):
#                 X_train_cv, X_valid_cv = self.X_train[train_index], self.X_train[val_index]
#                 y_train_cv, y_valid_cv = self.y_train[train_index], self.y_train[val_index]
#                 d_train_fold = xgboost.DMatrix(X_train_cv, y_train_cv)
#                 d_val_fold = xgboost.DMatrix(X_valid_cv, y_valid_cv)
#                 watchlist = [(d_train_fold,'train'),(d_val_fold,'eval')]
#
#                 model_fold = xgboost.train(xgb_params
#                                            , evals=watchlist
#                                            , verbose_eval=xgboost_verbose_eval
#                                            , early_stopping_rounds=early_stopping_rounds
#                                            , dtrain=d_train_fold
#                                            , num_boost_round=num_boost_round)
#                 S_train[val_index] = model_fold.predict(d_val_fold)
#                 score_tmp = score_metric(y_valid_cv, S_train[val_index])
#                 cv_scores.append(score_tmp)
#                 S_test_i[:, i] = model_fold.predict(d_test)
#                 print("Fold:{} --> score:{}.".format(i, score_tmp))
#
#             S_test = S_test_i.sum(axis=1) / n_folds
#             print("Mean:{}, Std:{}".format(np.mean(cv_scores), np.std(cv_scores)))
#             return S_train, S_test
#
#     def run_nn_stacker(self
#                        , build_model
#                        , epochs
#                        , batch_size
#                        , saved_path
#                        , saved_file_name
#                        , score_metric
#                        , patience = 5
#                        , decay_rate = 1
#                        , decay_after_n_epoch = 10
#                        , verbose_nn = 2
#                        ):
#         '''
#
#         Example:
#         -------
#         def build_model():
#             log = Sequential()
#             log.add(Dense(256, input_shape=(X_train.shape[1],), activation='relu'))
#             log.add(BatchNormalization())
#             log.add(Dropout(0.4))
#             log.add(Dense(128, activation='elu'))
#             log.add(BatchNormalization())
#             log.add(Dropout(0.3))
#             log.add(Dense(64, activation='elu'))
#             log.add(BatchNormalization())
#             log.add(Dropout(0.2))
#             log.add(Dense(32, activation='elu'))
#             log.add(BatchNormalization())
#             log.add(Dropout(0.1))
#             log.add(Dense(1, activation='sigmoid'))
#             log.compile(loss='binary_crossentropy', optimizer='adam') log must be compiled
#             return log
#         S_train, S_test = stack_generater.run_nn_stacker(build_model, 50, 128, PATH_DATA_MODEL, "nn_fold", 5, roc_auc_score, 0)
#         '''
#         with tick_tock("stacking", self.verbose):
#             n_folds = self.kf_n.n_splits
#             S_train = np.zeros_like(self.y_train).astype(np.float32)
#             S_test_i = np.zeros((self.X_test.shape[0], n_folds)).astype(np.float32)
#             cv_scores = []
#
#             for i, (train_index, val_index) in enumerate(self.kf_n.split(self.X_train, self.y_train)):
#
#                 X_train_cv, X_valid_cv = self.X_train[train_index], self.X_train[val_index]
#                 y_train_cv, y_valid_cv = self.y_train[train_index], self.y_train[val_index]
#
#                 model = build_model(X_train_cv.shape[1])
#                 callbacks_ins = callbacks_keras(saved_path + saved_file_name + '_v'+ str(i) + ".p"
#                                                 , model, patience=patience, decay_rate=decay_rate, decay_after_n_epoch=decay_after_n_epoch)
#                 model.fit(X_train_cv,y_train_cv, callbacks=callbacks_ins.callbacks
#                           ,validation_data=[X_valid_cv, y_valid_cv]
#                           ,epochs=epochs, batch_size=batch_size, verbose=verbose_nn)
#
#                 model.load_weights(saved_path + saved_file_name + '_v'+ str(i) + ".p")
#                 S_train[val_index] = np.squeeze(model.predict(X_valid_cv))
#                 score_tmp = score_metric(y_valid_cv, S_train[val_index])
#                 cv_scores.append(score_tmp)
#
#                 S_test_i[:, i] = np.squeeze(model.predict(self.X_test))
#                 print("Fold:{} --> score:{}.".format(i, score_tmp))
#
#             S_test = S_test_i.sum(axis=1) / n_folds
#             print("Mean:{}, Std:{}".format(np.mean(cv_scores), np.std(cv_scores)))
#             return S_train, S_test
#
#     def run_other_stackers(self, base_models, score_metric):
#         '''Stacking for sklearn mdoels
#            Parameters
#            ----------
#            base_models: models in list.
#                [model_1, model_2, ...], only support sklearn's models
#
#            Return
#            ------
#            S_train: numpy array
#                             stacked training data
#            S_test: numpy array
#                             stacked testing data
#         '''
#         with tick_tock("fitting stacking", self.verbose):
#             S_train = np.zeros((self.X_train.shape[0], len(base_models))).astype(np.float32)
#             S_test = np.zeros((self.X_test.shape[0], len(base_models))).astype(np.float32)
#             n_folds = self.kf_n.n_splits
#             cv_scores = []
#
#             print ("Fitting base models begin")
#             for i, model in enumerate(base_models):
#                 print ("Fitting the {0} th base models".format(i))
#                 S_test_i = np.zeros((self.X_test.shape[0], n_folds))
#                 for j, (train_index, val_index) in enumerate(self.kf_n.split(self.X_train, self.y_train)):
#                     X_train_cv, X_valid_cv = self.X_train[train_index], self.X_train[val_index]
#                     y_train_cv, y_valid_cv = self.y_train[train_index], self.y_train[val_index]
#
#                     model.fit(X_train_cv, y_train_cv)
#                     try:
#                         pred_valid = model.predict_proba(X_valid_cv)[:,1]
#                     except:
#                         pred_valid = model.predict(X_valid_cv)
#
#                     score_tmp = score_metric(y_valid_cv, pred_valid)
#                     cv_scores.append(score_tmp)
#
#                     S_train[val_index,i] = pred_valid
#                     S_test_i[:,j] = model.predict_proba(self.X_test)[:,1]
#
#                     print("Fold:{} --> score:{}.".format(j, score_tmp))
#                 S_test[:,i] = S_test_i.mean(1)
#
#             print("Mean:{}, Std:{}".format(np.mean(cv_scores)
#                                            , np.std(cv_scores)))
#             return S_train, S_test
#
#     def run_nn_embedding_stacker(self
#                        , build_model
#                        , epochs
#                        , batch_size
#                        , saved_path
#                        , saved_file_name
#                        , embedded_col_index
#                        , score_metric
#                        , patience = 5
#                        , decay_rate = 1
#                        , decay_after_n_epoch = 10
#                        , verbose_nn = 2
#                        ):
#         '''
#
#         Example:
#         -------
#         def build_model(embedded_info):
#             models = []
#
#             model_embedded_1 = Sequential()
#             model_embedded_1.add(Embedding(embedded_info[0], embedded_info[0]-1, input_length=1))
#             model_embedded_1.add(Reshape(target_shape=(embedded_info[0]-1,)))
#             models.append(model_embedded_1)
#
#             model_rest = Sequential()
#             model_rest.add(Dense(63, input_dim=64))
#             models.append(model_rest)
#
#             log = Sequential()
#             log.add(Merge(models, mode='concat'))
#             log.add(Dense(256, activation='relu'))
#             log.add(BatchNormalization())
#             log.add(Dropout(0.4))
#             log.add(Dense(128, activation='elu'))
#             log.add(BatchNormalization())
#             log.add(Dropout(0.3))
#             log.add(Dense(64, activation='elu'))
#             log.add(BatchNormalization())
#             log.add(Dropout(0.2))
#             log.add(Dense(32, activation='elu'))
#             log.add(BatchNormalization())
#             log.add(Dropout(0.1))
#             log.add(Dense(1, activation='sigmoid'))
#
#             log.compile(loss='binary_crossentropy', optimizer='adam') log must be compiled
#             return log
#         S_train, S_test = stack_generater.run_nn_stacker(build_model, 50, 128, PATH_DATA_MODEL, "nn_fold", 5, roc_auc_score, 0)
#         '''
#         with tick_tock("stacking", self.verbose):
#             n_folds = self.kf_n.n_splits
#             S_train = np.zeros_like(self.y_train).astype(np.float32)
#             S_test_i = np.zeros((self.X_test.shape[0], n_folds)).astype(np.float32)
#             cv_scores = []
#
#             for i, (train_index, val_index) in enumerate(self.kf_n.split(self.X_train, self.y_train)):
#
#                 X_train_cv, X_valid_cv = self.X_train[train_index], self.X_train[val_index]
#                 y_train_cv, y_valid_cv = self.y_train[train_index], self.y_train[val_index]
#
#                 X_train_list, X_valid_list, X_test_list, embedded_info = \
#                 self.prepare_embedding_data(X_train_cv, X_valid_cv, self.X_test, embedded_col_index)
#                 model = build_model(embedded_info)
#                 callbacks_ins = callbacks_keras(saved_path + saved_file_name + '_v'+ str(i) + ".p"
#                                                 , model, patience=patience, decay_rate=decay_rate, decay_after_n_epoch=decay_after_n_epoch)
#                 model.fit(X_train_list, y_train_cv, callbacks=callbacks_ins.callbacks
#                           ,validation_data=[X_valid_list, y_valid_cv]
#                           ,epochs=epochs, batch_size=batch_size, verbose=verbose_nn)
#
#                 model.load_weights(saved_path + saved_file_name + '_v'+ str(i) + ".p")
#                 S_train[val_index] = np.squeeze(model.predict(X_valid_list))
#                 score_tmp = score_metric(y_valid_cv, S_train[val_index])
#                 cv_scores.append(score_tmp)
#
#                 S_test_i[:, i] = np.squeeze(model.predict(X_test_list))
#                 print("Fold:{} --> score:{}.".format(i, score_tmp))
#
#             S_test = S_test_i.sum(axis=1) / n_folds
#             print("Mean:{}, Std:{}".format(np.mean(cv_scores)
#                                            , np.std(cv_scores)))
#             return S_train, S_test
#
#     @staticmethod
#     def prepare_embedding_data(X_train, X_valid, X_test, embedded_col_indexs):
#         '''
#             Parameters
#             ----------
#             X_train: numpy array
#             X_valid: numpy array
#             X_test: numpy array
#             embedded_cols: list
#                 col indexs going tobe embedd
#
#
#             Return
#             ------
#             embedded_info: return unique values of each embedding
#         '''
#         data = np.concatenate([X_train, X_valid, X_test])
#         normal_col_indexs = list(range(data.shape[1]))
#         for index in embedded_col_indexs:
#             normal_col_indexs.remove(index)
#
#
#         X_train_list = []
#         X_valid_list = []
#         X_test_list = []
#         embedded_info = dict()
#         for i, col in enumerate(embedded_col_indexs):
#             le = LabelEncoder()
#             le.fit(data[:,col])
#             tmp_train = le.transform(X_train[:,col])
#             tmp_valid = le.transform(X_valid[:,col])
#             tmp_test = le.transform(X_test[:,col])
#             X_train_list.append(tmp_train)
#             X_valid_list.append(tmp_valid)
#             X_test_list.append(tmp_test)
#             embedded_info[i] = le.classes_.shape[0]
#
#         X_train_list.append(X_train[:, normal_col_indexs])
#         X_valid_list.append(X_valid[:, normal_col_indexs])
#         X_test_list.append(X_test[:, normal_col_indexs])
#
#         return X_train_list, X_valid_list, X_test_list, embedded_info
#
# def ka_bagging_2class_or_reg(X_train, y_train, model, seed, bag_round
#                             , X_test, update_seed=True, is_classification=True, using_notebook=True):
#     '''
#         Bagging for "2-class classification" log and "regression" log
#
#         Parameters
#         ----------
#         X_train: numpy array 2-dimension
#              training data for fitting log
#         X_test: numpy array 2-dimension
#              testing data for predict result
#         y: numpy array 1-dimension
#              training target
#         model: log instance
#         seed: int
#              random seed
#         bag_round: int
#              bagging rounds
#         update_seed: boolean
#              update log to generate difference result
#         is_classification: boolean
#              classfication will predict probability by default
#              regression only predict value
#
#        Return:
#        baggedpred: numpy array
#              bagged prediction
#
#
#        Example
#        -------
#        Regression:
#            from sklearn.datasets import load_boston
#            from sklearn.metrics import mean_squared_error
#
#            data = load_boston()
#            X = data.data
#            y = data.target
#            X_test = X.copy()
#
#            log = RandomForestRegressor()
#            pred_10 = bagging_2class_or_reg(X, y, log, 10, 10, X_test, is_classification=False)
#            pred_1 = bagging_2class_or_reg(X, y, log, 10, 1, X_test, is_classification=False)
#
#            print(mean_squared_error(y, pred_10)) # 1.38465739328
#            print(mean_squared_error(y, pred_1)) # 2.13027490119
#
#        Classification:
#            from sklearn.datasets import load_boston
#            from sklearn.metrics import roc_auc_score
#
#            data = load_boston()
#            X = data.data
#            y = data.target
#            X_test = X.copy()
#
#            log = RandomForestRegressor()
#            pred_10 = bagging_2class_or_reg(X, y, log, 10, 10, X_test, is_classification=False)
#            pred_1 = bagging_2class_or_reg(X, y, log, 10, 1, X_test, is_classification=False)
#
#            print(mean_squared_error(y, pred_10)) # 0.998868778281
#            print(mean_squared_error(y, pred_1)) # 0.993778280543
#
#
#     '''
#     # create array object to hold predictions
#     baggedpred=np.zeros(shape=X_test.shape[0])
#     #loop for as many times as we want bags
#     if using_notebook:
#         for n in tqdm_notebook(range(0, bag_round)):
#             #shuffle first, aids in increasing variance and forces different results
#             X_train, y_train=shuffle(X_train, y_train, random_state=seed+n)
#
#             # update seed if requested, to give a slightly different log
#             # log like knn does not have random_state parameter
#             if update_seed:
#                 model.set_params(random_state=seed + n)
#
#             model.fit(X_train, y_train)
#             if is_classification:
#                 pred = model.predict_proba(X_test) # predict probabilities
#                 if pred.ndim == 1:
#                     pass
#                 elif pred.ndim == 2:
#                     pred = pred[:,1]
#                 else:
#                     print("this is a n>2 category problem, stacker is only suitable for 2-class and regression")
#             else:
#                 pred = model.predict(X_test)
#
#             baggedpred += pred/bag_round
#     else:
#         for n in tqdm(range(0, bag_round)):
#             #shuffle first, aids in increasing variance and forces different results
#             X_train, y_train=shuffle(X_train, y_train, random_state=seed+n)
#
#             # update seed if requested, to give a slightly different log
#             # log like knn does not have random_state parameter
#             if update_seed:
#                 model.set_params(random_state=seed + n)
#
#             model.fit(X_train, y_train)
#             if is_classification:
#                 pred = model.predict_proba(X_test) # predict probabilities
#                 if pred.ndim == 1:
#                     pass
#                 elif pred.ndim == 2:
#                     pred = pred[:,1]
#                 else:
#                     print("this is a n>2 category problem, stacker is only suitable for 2-class and regression")
#             else:
#                 pred = model.predict(X_test)
#
#             baggedpred += pred/bag_round
#     return baggedpred
#
# def ka_bagging_2class_or_reg_lgbm(X_train, y_train, seed, bag_round, params
#                                  , X_test, using_notebook=True, num_boost_round=0):
#     '''
#         early version
#     '''
#     # create array object to hold predictions
#     baggedpred=np.zeros(shape=X_test.shape[0]).astype(np.float32)
#     #loop for as many times as we want bags
#     if using_notebook:
#         for n in tqdm_notebook(range(0, bag_round)):
#             #shuffle first, aids in increasing variance and forces different results
#             X_train, y_train=shuffle(X_train, y_train, random_state=seed+n)
#             params['seed'] = seed + n
#             model = lightgbm.train(params, lightgbm.Dataset(X_train, y_train), num_boost_round=num_boost_round)
#             pred = model.predict(X_test)
#             baggedpred += pred/bag_round
#
#     return baggedpred



import os
import json
import time
import argparse
from itertools import chain
from datetime import datetime

import pandas as pd
from sklearn.externals import joblib
from IPython.display import display
from time import strftime, gmtime

#==============================================================================
# 时间相关
#==============================================================================
def log_start_time(fname):
    global st_time
    st_time = time()
    print("""
#=================================================================================
# START !!! {}    PID: {}    Time: {}
#=================================================================================
""".format(fname.split("/")[-1], os.getpid(), datetime.now().strftime('%Y-%m-%d %H:%M:%S')))

    #    send_line(f'START {fname}  time: {elapsed_minute():.2f}min')

    return

def log_end_time(fname):
    print("""
#=================================================================================
# SUCCESS !!! {}  Total Time: {} [s]
#=================================================================================
""".format(fname.split("/")[-1], _elapsed_minute()))
    return

def _elapsed_minute():
    return int(time() - st_time)

def now():
    return strftime("%Y-%m-%d %H:%M:%S", gmtime())
#==============================================================================
# 文件相关
#==============================================================================
def csv_2_pickle(paths: list):
    """Convert csv files into pickle format files.

     Parameters
     ----------
     paths: list
     	Csv file paths.

     Examples
     --------
     PATH = Path("data/raw/")
     CSV = [str(i) for i in list(PATH.glob("*.csv"))]
     csv_2_pickle(CSV)
     """
    for path in paths:
        data = pd.read_csv(path)
        joblib.dump(data, path.split('csv')[0] + "p")

def loadConfig(path):
    parser  = argparse.ArgumentParser()
    parser.add_argument('--config', default=path)
    options = parser.parse_args()
    config  = json.load(open(options.config))
    return config

def mkdir(path: str):
    """Create directory.

     Create directory if it is not exist, else do nothing.

     Parameters
     ----------
     path: str
     	Path of your directory.

     Examples
     --------
     mkdir("data/raw/train")
     """

    try:
        os.stat(path)
    except:
        os.mkdir(path)

def remove_temporary_files(folder_path: str):
    """Remove files begin with ".~".

     Parameters
     ----------
     folder_path: str
        Folder path which you want to clean.

     Examples
     --------
     remove_temporary_files("data/raw/")
     """
    '''
        Parameter
        remove files begin with ".~"
        ---------
        folder_path: str
    '''
    num_of_removed_file = 0
    for fname in os.listdir(folder_path):
        if fname.startswith("~") or fname.startswith("."):
            num_of_removed_file += 1
            os.remove(folder_path + '/'+ fname)
    print("{0} file have been removed".format(num_of_removed_file))
#==============================================================================
# 展示相关
#==============================================================================
def display_pro(data: pd.DataFrame, n=5):
    """Pro version of display function.

     Display [memory usage], [data shape] and [first n rows] of a pandas dataframe.

     Parameters
     ----------
     data: pandas dataframe
        Pandas dataframe to be displayed.
     n: int
        First n rows to be displayed.

     Example
     -------
     import pandas as pd
     from sklearn.datasets import load_boston
     data = load_boston()
     data = pd.DataFrame(data.data)
     display_pro(data)
     """

    '''
        Parameters
        ----------
        data: pandas dataframe


        Returns
        -------
            None
    '''
    memory = memory_usage(data, 0)
    print("Data shape   : {}".format(data.shape))
    display(data[:n])

def memory_usage(data: pd.DataFrame, detail=1):
    """Show memory usage.

     Parameters
     ----------
     data: pandas dataframe
     detail: int, optinal (default = 1)
        0: show memory of each column
        1: show total memory

     Examples
     --------
     import pandas as pd
     from sklearn.datasets import load_boston
     data = load_boston()
     data = pd.DataFrame(data.data)
     memory = memory_usage(data)
     """
    if detail:
        display(data.memory_usage())
    memory = data.memory_usage().sum() / (1024*1024)
    print("\nMemory usage : {0:.2f}MB".format(memory))
    return memory


#==============================================================================
# 功能相关
#==============================================================================
def find_a_not_in_b(a : list, b : list):
    '''找到a中存在b中不存在的元素

        Parameter
        ---------

        Return
        ------
        list
    '''
    # isinstance(a, float)是为了防止[空值]的情况
    if  isinstance(a, float) or isinstance(b, float):
        return a
    return [i1 for i1 in a if i1 not in b]

def find_a_in_b(a : list, b : list):
    '''找到a中存在b中也存在的元素

        Parameter
        ---------

        Return
        ------
        list
    '''
    # isinstance(a, float)是为了防止[空值]的情况
    if  isinstance(a, float) or isinstance(b, float):
        return []
    return [i1 for i1 in a if i1 in b]

def sum_list_in_list(list_in_list):
    '''将 list in list 变为 list
        Parameter
        ---------
        list_in_list: list

        Return
        ------
        list

        Example
        -------
        sum_list([['1','2'], ['2']]) -> ['1', '2', '2']
    '''
    return list(chain.from_iterable(list_in_list))

def unique_list_in_list(list_in_list):
    '''将 list in list 变为 list
        Parameter
        ---------
        list_in_list: list

        Return
        ------
        list

        Example
        -------
        unique_list_in_list([['1','2'], ['2']]) -> ['1', '2']
    '''
    li = sum_list_in_list(list_in_list)
    return list(set(li))

def is_primary_key(data, column_list):
    '''Verify if columns in column list can be treat as primary key

        Parameter
        ---------
        data: pandas dataframe

        column_list: list_like
                     column names in a list

        Return
        ------
        boolean: if true, these columns are unique in combination and can be used as a key
                 if false, these columns are not unique in combination and can not be used as a key
    '''

    return data.shape[0] == data.groupby(column_list).size().reset_index().shape[0]

def is_identical(col_1, col_2):
    '''判断数据集的两列是否一致

        Parameters
        ----------
        data: dataframe
        col_1: string
        col_2: string

        Return
        ------
        True、False
    '''
    if sum(col_1 == col_2) / len(col_1) == 1:
        return True
    return False

def reverse_dict(x : dict):
    '''翻转一个dict

        Parameters
        ----------
        x: dict

        Return
        ------
        dict
    '''
    return {v:k for k, v in x.items()}

def min_max(x):
    return min(x), max(x)



import time
import json
import argparse

from itertools               import islice

def loadConfig(path):
    parser  = argparse.ArgumentParser()
    parser.add_argument('--config', default=path)
    options = parser.parse_args()
    config  = json.load(open(options.config))
    return config

class tick_tock:
    def __init__(self, process_name, verbose=1):
        self.process_name = process_name
        self.verbose = verbose
    def __enter__(self):
        if self.verbose:
            print(self.process_name + " begin ......")
            self.begin_time = time.time()
    def __exit__(self, type, value, traceback):
        if self.verbose:
            end_time = time.time()
            print(self.process_name + " end ......")
            print('time lapsing {0} s \n'.format(end_time - self.begin_time))

# class callbacks_keras:
#     def __init__(self, filepath, model,
#                  base_lr=1e-3, decay_rate=1,
#                  decay_after_n_epoch=10, patience=20,
#                  mode='min', monitor='val_loss'):
#         self.base_lr = base_lr
#         self.model = model
#         self.decay_rate = decay_rate
#         self.decay_after_n_epoch = decay_after_n_epoch
#         self.callbacks = [ModelCheckpoint(filepath = filepath,
#                                           monitor = monitor,
#                                           verbose = 2,
#                                           save_best_only = True,
#                                           save_weights_only = True,
#                                           mode = mode),
#                          EarlyStopping(monitor = monitor, patience = patience, verbose=2, mode = mode),
#                          LearningRateScheduler(self._scheduler)]
#
#     def _scheduler(self, epoch):
#         if epoch%self.decay_after_n_epoch==0 and epoch!=0:
#             lr = K.get_value(self.model.optimizer.lr)
#             K.set_value(self.model.optimizer.lr, lr*self.decay_rate)
#             print("lr changed to {}".format(lr*self.decay_rate))
#         return K.get_value(self.model.optimizer.lr)
#
# def ka_xgb_r2_error(preds, dtrain):
#     labels = dtrain.get_label()
#     return 'error', r2_score(labels, preds)
#
# def ka_xgb_r2_exp_error(preds, dtrain):
#     labels = dtrain.get_label()
#     preds = np.clip(np.exp(preds),0, 1e10)
#     return 'error', r2_score(np.exp(labels), preds)
#
# def ka_erfinv_rank_transform(x):
#     '''
#         This is used on numeric variable, after doing this operation, one should do MM and SS on all dataset.
#     '''
#     mm = MinMaxScaler()
#     tmp = erfinv(np.clip(np.squeeze(mm.fit_transform(rankdata(x).reshape(-1,1))), 0, 0.999999999))
#     tmp = tmp - np.mean(tmp)
#     return tmp
#
# def kaggle_points(n_teams, n_teammates, rank, t=1):
#     return (100000 / np.sqrt(n_teammates)) * (rank ** (-0.75)) * (np.log10(1 + np.log10(n_teams))) * (np.e**(t/500))
######################################################################################################
# read and write file functions
######################################################################################################
def ka_dict_head(dic, n):
    '''
        first n rows of a dictionary
    '''
    return list(islice(dic, n))

def ka_sort_dict(a, by_key=False, reverse=False):
    '''
        sort dictionary by item or key.
    '''
    sorted_x = sorted(a.items(), key=operator.itemgetter(1-by_key), reverse=reverse)
    return sorted_x



from __future__ import absolute_import
import sys
import os
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(name = "kagglebuddy",
                 version = "0.0.18",
                 author = "鲲(China)",
                 author_email = "972775099@qq.com",
                 description = "machine learning toolkit",
                 long_description = long_description,
                 long_description_content_type = "text/markdown",
                 url = "https://github.com/NickYi1990/kaggleBuddy",
                 packages = setuptools.find_packages(),
                 classifiers=[
                              "Programming Language :: Python :: 3",
                              "License :: OSI Approved :: MIT License",
                              "Operating System :: OS Independent",
                              ])


# if __name__ == "__main__":
#     CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))
#     sys.path.insert(0, CURRENT_DIR)
#
#     with open("README.md", "r") as fh:
#         long_description = fh.read()
#
#     setuptools.setup(
#                      name                          = "kagglebuddy",
#                      version                       = "0.0.10",
#                      author                        = "鲲(China)",
#                      author_email                  = "972775099@qq.com",
#                      description                   = "machine learning toolkit",
#                      long_description              = long_description,
#                      long_description_content_type = "text/markdown",
#                      url                           = "https://github.com/NickYi1990/kaggleBuddy",
#                      packages                      = setuptools.find_packages(),
#                      classifiers=[
#                                   "Programming Language :: Python :: 3",
#                                   "License :: OSI Approved :: MIT License",
#                                   "Operating System :: OS Independent",
#                                   ],
#                      )
